A helper script for rendering paper figures with Blender

## Quick Usage

```
python render/render_script.py --overrides mesh.path=path/to/mesh.obj
```

This library uses `omegaconf` to manage configs using YAML. You can set args via the command line, or config yaml files. Multiple YAML files can be passed, each overriding the previous (think CSS). This is useful for saving shared settings for groups of renders.

The default configuration is stored in `meshgen-tools/render/render_config_base.yml`. See there for many options supported by the renderer, with some documentation comments.

## Installing

This rendering tool is a Blender python script. It is possible to run python scripts via any installation of Blender, however then you are stuck using the version of Python that come with Blender, and it is difficult to install/use additional packages, etc.

Instead, we will use the new `bpy` Blender Python packages, which are hosted precompiled on PyPi. The package comes with its own binaries of Blender bundled as a part of the the precompiled wheel. The `bpy` package is slightly weird: each version of Blender supports exactly one version of Python. This library was tested with Python 3.10 / Blender 4.0, though others may work. If something is mismatched or unsupported about your platform, `pip` may instead try to fetch an ancient `2.XX` version of Blender (see Gotchas below).

```sh
# clone this repo
git clone --recursive ssh://git@gitlab-master.nvidia.com:12051/nsharp/meshgen-tools.git

# create a conda env
conda create -n "blender" python=3.10 
conda activate blender

# install standard packages
pip install bpy numpy scipy pillow matplotlib omegaconf libigl

# render something
cd render

```

## More Usage

### Experiment rendering workflow with omegaconf

Suppose you are creating a figure where a few meshes need to be rendered in the same style. I suggest creating a subdirectory like
```sh
experiments/
  this_experiment/
    experiment_style.yaml
    render_mesh1.yaml
    render_mesh2.yaml
    mesh_1.obj
    mesh_2.obj
```

The `experiment_style.yaml` contains shared settings for all images for this experiment (like mesh color, wireframes, etc) 
```yaml
# experiment_style.yaml
render:
  res_x: 1200
  res_y: 1200
  threshold_ground_shadows: True
  threshold_ground_shadows_alpha: 0.02

scene:
  transform_translate: 'auto'
  transform_scale: 'auto'
  with_ground_shadows: True
  ground_level: 'auto'

mesh:
  color: 'darkred'
  wireframe_tubes: True
  wireframe_color: 'black'
```

Then the per-image yaml files specify the input and output data, and any values which must be customized per-input (such as nicely positioning the camera)
```yaml
# render_mesh1.yaml
mesh:
  path: mesh1.obj
output:
  image_filename: 'mesh1_render.png'
view:
  camera_location: [-2.3, 3., 2.5]
  camera_look_at: [0., 0., 0.25]
```

Then to render, run 
```sh
python render/render_script.py --configs experiment_style.yaml render_mesh1.yaml
python render/render_script.py --configs experiment_style.yaml render_mesh2.yaml
```
which can easily be modified and reproduced to regenerate figures later as needed.


### More examples

(many of these options appear in the default configuration, see `render_config_base.yml`)

Set output paths, render a not-transparent image on a white background
```
python render/render_script.py --overrides mesh.path=path/to/mesh.obj outputs.image_filename=my_render.png render.transparent=False render.background_color=FFFFFF
```

Adjust rendering resolution and performance
```
python render/render_script.py --overrides mesh.path=path/to/mesh.obj outputs.image_filename=my_render.png render.res_x=1000 render.res_y=800 render.num_samples=128 render.exposure=1.7
```

Add a transparent ground plane to catch shadows, positioned automatically
```
python render/render_script.py --overrides mesh.path=path/to/mesh.obj outputs.image_filename=my_render.png scene.with_ground_shadows=True scene.ground_level=auto  render.threshold_ground_shadows=True
```

Position meshes and cameras (check `.yaml` for camera options, it is hard to enter vectors at the command line)
```
python render/render_script.py --overrides mesh.path=path/to/mesh/mesh.obj scene.transform_scale=0.01 scene.transform_translate=auto scene.ground_level=-0.25
```

Handle Z-up (vs. Y-up) data
```
python render/render_script.py --overrides mesh.path=path/to/mesh/mesh.obj mesh.up_axis=Z
```

Set mesh color, add a nice edge wireframe
```
python render/render_script.py --overrides mesh.path=path/to/mesh.obj mesh.color=F2B79A mesh.wireframe_tubes=True mesh_appearance.wireframe_color='darkred'
```

You an specify colors in many different ways:
  - `[0.3, 0.5, 0.5]` RGB floats on `[0,1]`
  - `[0.3, 0.5, 0.5, 1.0]`
  - `'#AB2F3D'`
  - `'AB2F3D'` (alternate way to specify hex, because # is hard to type on the command line)
  - `'unthreatening_blue'`  our custom color names, see dict in source file
  - `'darkgreen'`  # any matplotlib color name https://matplotlib.org/3.2.1/gallery/color/named_colors.html
  - `'xkcd:kermit green'` named colors from the xkcd color survey https://xkcd.com/color/rgb/

All of these can be entered in a YAML file. Some of these may be hard to type in on the terminal, but hex colors and string names are easy.

Render a scalar function along the mesh. The scalar will be read from the `u` channel of the UV map in the file. For the colormap, you can pass any matplotlib colormap name.
```
python render/render_script.py --overrides mesh.path=path/to/mesh.obj mesh.scalar_func_source='uv_u' mesh.up_axis=Y mesh.forward_axis=Z mesh.scalar_func_vmin=0. mesh.scalar_func_vmax=0.4 mesh.scalar_func_cmap='Reds'
```

### Rendering multiple meshes / point clouds / etc into the scene at once

You can add any number of mesh etc entries under the `content` list. They will all be added into the resulting scene. Each of these entries can take all of the same options as when they appear at the top level of the file.

**Example:**
```yaml
content:
  - mesh:
      path: /home/nsharp/mesh/spot.obj
      up_axis: 'Y'
      forward_axis: 'NEGATIVE_Z'
      color: 'green'
      shade_smooth: True

  - mesh:
      path: /home/nsharp/mesh/bunny.obj
      name: rabbit2
      up_axis: 'Y'
      forward_axis: 'NEGATIVE_Z'
      color: 'purple'
      wireframe_tubes: True
      shade_smooth: True
      transform_translate: [1., 2., 3.]
      transform_scale: 0.2

  - point_cloud:
      path: /home/nsharp/mesh/armadillo.obj
      up_axis: 'Z'
      forward_axis: 'X'
      transform_translate: [0., 0., 0.]
      transform_scale: 1.0
      radius: 0.005
      color: 'grey'
```

### Batch rendering and animations

If instead of a path to a file, you pass a path to a directoy, the script will loop over all files in the the directory and render them
```sh
python render/render_script.py --overrides mesh.path=path/to/mesh_dir
```

Render a batch of frames which can be encoded into a movie. (Right now this just changes a few default behaviors)
```sh
python render/render_script.py --overrides mesh.path=path/to/mesh_dir output.movie=True
```

Render a turntable animation
```sh
python render/render_script.py --overrides mesh.path=~/path/to/mesh.obj render.transparent=False render.animate_frames=120 render.animate_camera='turntable' render.animate_camera_turntable_period=120
```

A good ffmpeg command to encode the frames into a movie is:
```sh
ffmpeg -i render_%06d.png -c:v libx264 -crf 18 -profile:v main -pix_fmt yuv420p  -c:a aac -ac 2 -b:a 128k -movflags faststart output.mp4
```

### All mesh settings

An exhaustive list of mesh settings, with some inline comments

```yml
mesh:

  # Mesh to load
  path: '/home/nsharp/mesh/cad/bracket.obj'

  # Up-direction conventions for the mesh
  # (use this in lieu of a full rotation option)
  up_axis: 'Y'
  forward_axis: 'NEGATIVE_Z'
     
  # Apply a custom transformation to the mesh
  # (note that there are separate scene-wide options in `scene` above, 
  # including to auto-center/scale the whole scene)
  transform_translate: [0., 0., 0.]
  transform_scale: 1.

  # Optionally hide the mesh from actually being rendered
  # (useful e.g. if you just want to draw a wireframe)
  show_mesh: True
  
  # call Blender's validate(), which fixes invalid inds, repeated faces, etc
  blender_validate_cleanup: True 
 
  # how do polygons get triangulated for rendering? by default we let
  # blender handle it, but alternate we can use our own tesselated from
  # the meshgen_tools package (which must be installed)
  triangulate_polygons_with: 'BLENDER' # alternately: 'MESHGEN_TOOLS'
  
  #### For this and all other color arguments, colors can be specified in any of the following ways:
  #   color: [0.3, 0.5, 0.5]
  #   color: [0.3, 0.5, 0.5, 1.0]
  #   color: '#AB2F3D'
  #   color: 'AB2F3D' # (alternate way to specify hex, because # is hard to type on the command line)
  #   color: 'unthreatening_blue' # our custom color names, see dict in source file
  #   color: 'darkgreen'  # any matplotlib color name https://matplotlib.org/3.2.1/gallery/color/named_colors.html
  #   color: 'xkcd:kermit green' # named colors from the xkcd color survey https://xkcd.com/color/rgb/
  color: 'unthreatening_blue'

  # Smooth vs. flat surface shading 
  shade_smooth: False 

  # Render edges as wireframe tubes
  wireframe_tubes: True
  wireframe_color: 'white'
  wireframe_tubes_radius: 0.0025
  wireframe_tubes_vertex_radius_mult: 1.5

  # Render a colormapped scalar function along the surface
  scalar_func_source: False # 'uv_u' to read from U channel of UVs
  scalar_func_vmin: 0.
  scalar_func_vmax: 1.
  scalar_func_cmap: 'viridis'
```

### All point cloud settings

An exhaustive list of mesh settings, with some inline comments.

```yml
point_cloud:
  # pass any mesh format, its vertices will be treated as the points
  path: /home/nsharp/mesh/armadillo.obj
  up_axis: 'Z'
  forward_axis: 'X'
  transform_translate: [0., 0., 0.]
  transform_scale: 1.0
  radius: 0.005
  color: 'grey'
```

## FAQs / Gotchas

### I am getting an ancient version of Blender when I install `bpy`

Something is mismatched about your platform/Python, so it is not matching the precompiled wheel. Are you using Python 3.10? We have also seen a more subtle version of this problem due to unsupported GLIBC verions.

### Renders are slow

Rendering should use your GPU and take just a few seconds. If it takes more than that, something is wrong, most likely that Blender is not able to use your GPU. The easiest way to confirm if your GPU is being used it to check `nvidia-smi` while rendering. If it is not being used, try `render.GPU_compute_device='CUDA'` (instead of the default 'OPTIX').
