import sys, os, fnmatch, re
import argparse

import bpy
import mathutils # available after importing bpy

from omegaconf import OmegaConf
import numpy as np
from PIL import Image
from PIL import ImageColor
import matplotlib
from matplotlib import colors as mcolors
import matplotlib.cm
import igl

'''
INSTRUCTIONS

Install the blender `bpy` package. Note that it is a a bit finnicky about version; each version of Blender has exactly one version of python which it is compatible with. This was tested with Blender 4.0 / Python 3.10. Others may work too.
'''

CWD = os.getcwd()
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
TMP_DIR = os.path.join(CWD, 'tmp_renderfiles')
DEFAULT_CONFIG_PATH = os.path.join(SCRIPT_DIR, 'render_config_base.yaml')
EXTRAS_PATH = os.path.join(SCRIPT_DIR, 'extras.blend')

DEFAULT_MESH_CONFIG_STR = """
name: ???
up_axis: 'Z'
forward_axis: 'X'
transform_translate: [0., 0., 0.]
transform_scale: 1.0
show_mesh: True
blender_validate_cleanup: True
triangulate_polygons_with: 'BLENDER' # or 'MESHGEN_TOOLS'
color: 'grey'
shade_smooth: False 
wireframe_tubes: False
wireframe_color: 'white'
wireframe_tubes_radius: 0.0025
wireframe_tubes_vertex_radius_mult: 1.5
scalar_func_source: null # 'uv_u'
scalar_func_vmin: 0.
scalar_func_vmax: 1.
scalar_func_cmap: 'viridis'
"""
DEFAULT_MESH_CONFIG = OmegaConf.create(DEFAULT_MESH_CONFIG_STR)

DEFAULT_POINT_CLOUD_CONFIG_STR = """
name: ???
up_axis: 'Z'
forward_axis: 'X'
transform_translate: [0., 0., 0.]
transform_scale: 1.0
radius: 0.005
color: 'grey'
"""
DEFAULT_POINT_CLOUD_CONFIG = OmegaConf.create(DEFAULT_POINT_CLOUD_CONFIG_STR)

def split_filepath(filepath):
    directory, filepath = os.path.split(filepath)
    base, extension = os.path.splitext(filepath)
    return directory, base, extension

def ensure_dir_exists(dir):
    try:
        os.makedirs(dir)
    except FileExistsError:
        pass

our_named_colors = {
    'unthreatening_blue' : (.098, .349, .978)
}
matplotlib_named_colors = dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS, **mcolors.TABLEAU_COLORS, **mcolors.XKCD_COLORS)

def is_hex_color(color_string):
    # accepts strings like '#AB02FF' or '#A40' (possibly with the # omitted)
    if not isinstance(color_string, str):
        return False
    hex_color_pattern = r'^#?([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$'
    return bool(re.match(hex_color_pattern, color_string))

def color_to_01_4tuple(color):

    if is_hex_color(color):
        if not color.startswith('#'): # it's hard to type #FFFFFF on the command line, so you can omit the # and we will add it
            color = '#' + color
        color = ImageColor.getcolor(color, "RGB")

    color = tuple(color)

    # if all int, convert to float / 255.
    all_int = all([isinstance(x, int) for x in color])
    if all_int:
        color = tuple([x / 255. for x in color])

    # if length 3, padd to length 4
    if len(color) == 3:
        color = (color[0], color[1], color[2], 1.0)

    for c in color:
        if c > 1.0:
            raise ValueError(f"color tuple has value {c}, which is >1.0. This is probably a bug, colors should be on [0,1]")

    return color


def resolve_color(arg):
    print(f"resolving color {arg}")

    # Try to access it as a tuple
    if not isinstance(arg, str):
        argtup = tuple(arg)
        return color_to_01_4tuple(argtup)
    
    if is_hex_color(arg):
        return color_to_01_4tuple(arg)

    # Check if it is one of our named colors
    if isinstance(arg, str):
        arg = arg.lower()
        if arg in our_named_colors:
            return color_to_01_4tuple(our_named_colors[arg])
        if arg in matplotlib_named_colors:
            return color_to_01_4tuple(matplotlib_named_colors[arg])
    
    raise ValueError(f"Could not resolve color: {arg}")

def save_matplotlib_colormap_as_tmp_image_file(cmap_name):
    res=(10,512)

    ensure_dir_exists(TMP_DIR)
    outfile_path = os.path.join(TMP_DIR, f"tmp-colormatp-{cmap_name}.png")

    cmap = matplotlib.colormaps[cmap_name]
    vals = np.linspace(0,1,num=res[1])
    colors = cmap(vals)

    colors = np.broadcast_to(colors[None,:,:], (res[0], res[1], 4))
    image_data = (colors * 255).astype(np.uint8)
    image = Image.fromarray(image_data)
    image.save(outfile_path)

    return outfile_path

def compute_mesh_scene_stats(global_cfg, mesh_cfg):

    # Read the mesh then throw it away
    tmp_mesh, verts, faces = read_mesh(global_cfg, mesh_cfg, return_verts_and_faces=True)
    bpy.data.objects.remove(tmp_mesh)

    bbox_min = np.min(verts, axis=0)
    bbox_max = np.max(verts, axis=0)

    bbox_center = 0.5 * (bbox_min + bbox_max)

    # Compute bounding box diagonal
    bbox_diag = np.linalg.norm(bbox_max - bbox_min)

    # find the min
    lowest_coord = np.min(verts[:,2])

    return {
        # 'lowest_coord' : lowest_coord, # unused currently
        # 'scale' : bbox_diag,
        # 'center' : bbox_center,
        'bbox_min' : bbox_min,
        'bbox_max' : bbox_max,
    }
         
def compute_point_cloud_scene_stats(global_cfg, point_cloud_cfg):
    # for now, these are identical
    return compute_mesh_scene_stats(global_cfg, point_cloud_cfg)

def generate_lines_obj(verts, faces):

    # Write out a line obj file
    ensure_dir_exists(TMP_DIR)
    outfile_path = os.path.join(TMP_DIR, "tmp-renderlines.obj")
    with open(outfile_path, 'w') as outfile:

        # this is a unique set of the actual vertex coords, not the inds
        edge_verts_set = set()

        for face in faces:
            D = len(face)
            for i in range(D):
                j = (i + 1) % D
                vi = verts[face[i],:]
                vj = verts[face[j],:]
                # again, this is a lexicographic sort of the actual
                # coords, not the inds
                edge_tup = tuple(sorted((tuple(vi), tuple(vj))))
                edge_verts_set.add((edge_tup))


        # Write out the actual file
        for v1, v2 in edge_verts_set:
            outfile.write(f"v {v1[0]} {v1[1]} {v1[2]}\n")
            outfile.write(f"v {v2[0]} {v2[1]} {v2[2]}\n")
        iV = 0
        for v1, v2 in edge_verts_set:
            outfile.write(f"l {iV + 1} {iV + 2}\n")
            iV += 2

    return outfile_path

def composite_over(image_in_path, background_color, image_out_path=None):
    img = Image.open(image_in_path)
    img = img.convert("RGBA")
    background = Image.new('RGBA', img.size, background_color)
    composite_image = Image.alpha_composite(background, img)
    composite_image = composite_image.convert('RGB')
    if image_out_path is None:
        image_out_path = image_in_path
    composite_image = composite_image.convert('RGB')
    composite_image.save(image_out_path)

def initialize_blender(cfg):
    # adapted from BlenderToolbox

    # clear out the default scene
    bpy.ops.wm.read_homefile()
    bpy.ops.object.select_all(action = 'SELECT')
    bpy.ops.object.delete() 

    # use the cycles renderer
    bpy.context.scene.render.engine = 'CYCLES'
    bpy.context.scene.render.resolution_x = cfg.render.res_x
    bpy.context.scene.render.resolution_y = cfg.render.res_y
    bpy.context.scene.render.resolution_percentage = cfg.render.res_percentage
    bpy.context.scene.cycles.samples = cfg.render.num_samples 
    bpy.context.scene.cycles.max_bounces = 6
    bpy.context.scene.render.film_transparent = True # we ALWAYS use transparent here, for non-transparent settings we composite below

    bpy.context.scene.view_settings.view_transform = "Filmic"
    bpy.context.scene.view_settings.look = "Medium High Contrast"
    bpy.context.scene.cycles.film_exposure = cfg.render.exposure

    if cfg.render.use_GPU:
        # Added by nsharp:
        # For me, the blender toolbox use_GPU=True setting was not sufficient to make blender use the GPU.
        # Some combination of these options makes it render on the GPU, I am not sure which settings are necessary or why
        preferences = bpy.context.preferences
        cycles_preferences = preferences.addons['cycles'].preferences
        cycles_preferences.compute_device_type = cfg.render.GPU_compute_device
        bpy.context.scene.cycles.device = 'GPU'
        cycles_preferences.get_devices()
        bpy.context.preferences.addons['cycles'].preferences.devices[0].use = True 


    # Denoising
    # WARNING: denoising like this kills the alpha shadow channel. 
    # instead we use a compositor graph
    denoise = cfg.render.denoise and (not cfg.scene.with_ground_shadows)
    bpy.data.scenes[0].view_layers[0]['cycles']['use_denoising'] = 1 if denoise else 0

def print_all_scene_object_names():
    # Useful for debugging
    print("\n=== All Objects Names:")
    for collection in bpy.data.collections:
       print(collection.name)
       for obj in collection.all_objects:
          print(f"  {obj.name}")

def delete_object(obj):
    bpy.ops.object.select_all(action='DESELECT')

    # special handling for some types (sorry, world)
    if isinstance(obj, bpy.types.Image):
        bpy.data.images.remove(obj)

    else: # everything else
        obj.select_set(True)
        bpy.ops.object.delete()

def resolve_possible_batch_rendering_path(global_cfg, obj_cfg):
    # Returns a tuple of
    #     - the exact path to the file to load for the current rendering iteration
    #     - max number of files that this will resolve to

    if os.path.isfile(obj_cfg.path):
        return obj_cfg.path, 1
    
    if os.path.isdir(obj_cfg.path):

        # assemble a sorted list of files in the tiletypes we support
        content_files = []
        for file in os.listdir(obj_cfg.path):
            if fnmatch.fnmatch(file, '*.obj') or fnmatch.fnmatch(file, '*.ply'):
                content_files.append(file)
  
        if len(content_files) == 0:
            raise ValueError(f"directory does not contain any content files {obj_cfg.path}")
        
        content_files.sort()

        resolved_file = os.path.join(obj_cfg.path, content_files[global_cfg.batch_render_ind])

        return resolved_file, len(content_files)

    # TODO support patterns a la FFMPEG?

    raise ValueError(f"path is neither flile nor directory {obj_cfg.path}")

def create_mesh_from_buffers(name, verts, faces, validate=True):
    # verts, faces can be any python iterable of positions / face indices,
    # such as numpy arrays or a nested list of faces 

    mesh_data = bpy.data.meshes.new(name=name)
    mesh_data.from_pydata(verts, [], faces)
    mesh_data.update()

    if validate:
        had_effect = mesh_data.validate()
        if had_effect:
            print("================================")
            print("WARNING: blender validate() changed the mesh, something may have been degenerate/broken in create_mesh_from_buffers()")
            print("================================")

    mesh_object = bpy.data.objects.new(name, mesh_data)

    bpy.context.scene.collection.objects.link(mesh_object)
    bpy.context.view_layer.update()

    return mesh_object 


def read_mesh(global_cfg, cfg, location=(0., 0., 0.), scale=(1., 1., 1.), name=None, return_verts_and_faces=False, use_triangulated=False):

    # Get the path
    # warning: path logic duplicated in meshgen_triangulate()
    if use_triangulated:
        path = cfg.path_triangulated 
    else:
        path, _ = resolve_possible_batch_rendering_path(global_cfg, cfg)

    if not os.path.isfile(path):
        raise ValueError(f"File does not exist: {path}")

    print(f"Loading mesh: {path}")

    # Do some sanity checking
    if hasattr(cfg, 'up_dir'):
        raise ValueError("You used the attribute 'up_dir', when it should be 'up_axis'")

    # Build a list of all objects before, to find the new one
    prev = []
    for ii in range(len(list(bpy.data.objects))):
        prev.append(bpy.data.objects[ii].name)

    _, extension = os.path.splitext(path)
    if extension == '.ply' or extension == '.PLY':
        bpy.ops.wm.ply_import(filepath=path, up_axis=cfg.up_axis, forward_axis=cfg.forward_axis)
    elif extension == '.obj' or extension == '.OBJ':
        bpy.ops.wm.obj_import(filepath=path, up_axis=cfg.up_axis, forward_axis=cfg.forward_axis, use_split_objects=False, use_split_groups=False)
    else:
        raise TypeError("read_mesh only supports .ply and .obj")

    # Build a list of all objects after, to find the new one
    after = []
    for ii in range(len(list(bpy.data.objects))):
        after.append(bpy.data.objects[ii].name)

    name = list(set(after) - set(prev))[0] # find the new one
    mesh = bpy.data.objects[name]
        
    print(f"   #verts = {len(mesh.data.vertices)}")

    if 'blender_validate_cleanup' in cfg and cfg.blender_validate_cleanup:

        # should we warn here? I guess not, validate() shouldn't change the image we make
        # if cfg.triangulate_polygons_with != 'BLENDER':
        #     raise ValueError("you probably do not want to use a custom tessellator AND the Blender validate() repair, there may be a risk that validate() disturbs the mesh. Try setting `mesh.blender_validate_cleanup: False`.")

        had_effect = mesh.data.validate()
        if had_effect:
            print("================================")
            print("WARNING: blender validate() changed the mesh, something may have been degenerate/broken in your input mesh")
            print("================================")
   
    # select this mesh
    bpy.context.view_layer.objects.active = mesh
    
    # First apply any local transformations
    mesh.location = cfg.transform_translate
    mesh.scale = (cfg.transform_scale, cfg.transform_scale, cfg.transform_scale)
    bpy.ops.object.transform_apply(location=True, rotation=True, scale=True)

    # Then apply any additional transformations
    mesh.location = location
    mesh.scale = scale
    bpy.ops.object.transform_apply(location=True, rotation=True, scale=True)

    bpy.context.view_layer.update()

    if return_verts_and_faces:
        obj_data = mesh.to_mesh()
        verts = [(v.co[0], v.co[1], v.co[2]) for v in obj_data.vertices]
        verts = np.array(verts)
        faces = [[int(i) for i in poly.vertices] for poly in obj_data.polygons]
        return mesh, verts, faces

    else:
        return mesh 

def meshgen_triangulate(global_cfg, cfg):
    import meshgen_tools

    # Get the path
    # warning: path logic duplicated in read_mesh()
    path, _ = resolve_possible_batch_rendering_path(global_cfg, cfg)
    if not os.path.isfile(path):
        raise ValueError(f"File does not exist: {path}")

    print(f"Custom-triangulating mesh: {path}")

    if cfg.scalar_func_source is not None:
        raise ValueError('retriangulator does not preseve scalar functions')
    
    directory, base, extension = split_filepath(path)
    out_file_path = os.path.join(TMP_DIR, f"{base}-triangulate.obj")

    # load the mesh 
    V, F_flat, F_deg = meshgen_tools.load_mesh_poly_flat(cfg.path)

    # apply our triangulator
    F_tri = meshgen_tools.triangulate_polygons_flat_list(V, F_flat, F_deg)

    # save hte result
    igl.write_obj(out_file_path, V, F_tri)

    cfg.path_triangulated = out_file_path 

def configure_mesh_in_scene(global_cfg, mesh_cfg):
    created_objects = []


    if mesh_cfg.triangulate_polygons_with == 'MESHGEN_TOOLS':
        # load and re-save an alternate triangulated version of the mesh
        # we will re-load this at the very end to render
        # (but no sooner, because we still want wireframes etc to come from the untriangulated mesh)
        meshgen_triangulate(global_cfg, mesh_cfg)    

    # global transforms
    # (local transforms are stored inside mesh_cfg and separately applied internally in read_mesh())
    global_translate = global_cfg.scene.global_transform_translate
    s = global_cfg.scene.global_transform_scale
    global_scale = (s, s, s) 

    ## read main mesh
    need_verts = mesh_cfg.wireframe_tubes
    if need_verts:
        mesh, verts, faces = read_mesh(global_cfg, mesh_cfg, global_translate, global_scale, return_verts_and_faces=True)
    else:
        mesh = read_mesh(global_cfg, mesh_cfg, global_translate, global_scale, return_verts_and_faces=False)
    
    # make sure the names in the config and the blender object match
    # (if no name was given in the yml, use Blender's auto-assigned name)
    if 'name' not in mesh_cfg:
        mesh_cfg.name = mesh.name
    else:
        mesh.name = mesh_cfg.name
    
    ## Generate wireframe tubes, if requested
    if mesh_cfg.wireframe_tubes:
    
        ## Pre-process an obj mesh to construct an tmp file which will load edge wireframe line objects of that mesh
        line_obj_file = generate_lines_obj(verts, faces)
        
        # create curve objects as the lines by reading from the temporary file we
        # prepared previously
        lines_cfg = OmegaConf.merge( # the obj above was written in blender coords
                DEFAULT_MESH_CONFIG,
                OmegaConf.create({
                    'path' : line_obj_file, 
                    'up_axis' : 'Z', 
                    'forward_axis' : 'Y'
                })
            )
        lines_object = read_mesh(global_cfg, lines_cfg)
        created_objects.append(lines_object)
        lines_object.name = f"{mesh_cfg.name}-WireframeCurves"
        bpy.ops.object.convert(target='CURVE')
        lines_object.data.bevel_resolution = 16
        lines_object.data.bevel_depth = mesh_cfg.wireframe_tubes_radius
        bpy.context.view_layer.objects.active = lines_object
        bpy.ops.object.shade_smooth()

        wireframe_mat = create_material(resolve_color(mesh_cfg.wireframe_color), f"{mesh_cfg.name}-Wireframe-MeshMat")
        assign_mesh_material(lines_object, wireframe_mat)

        # create a point cloud of the mesh vertices
        # first read the mesh again, this one will get geomety-nodes-ified into points
        mesh_points = read_mesh(global_cfg, mesh_cfg, global_translate, global_scale)
        mesh_points.name = f"{mesh_cfg.name}-WireframePoints"
        created_objects.append(mesh_points)


        # Now add a Geometry Node modifier:
        # https://blender.stackexchange.com/questions/316960/how-to-add-geometry-nodes-using-python
        modifier = mesh_points.modifiers.new("ScriptAdded", "NODES")

        node_group = bpy.data.node_groups.new("group name", type='GeometryNodeTree')
        node_group.interface.new_socket(name="Geometry In", in_out ="INPUT", socket_type="NodeSocketGeometry")
        node_group.interface.new_socket(name="Geometry Out", in_out ="OUTPUT", socket_type="NodeSocketGeometry")

        node_in        = node_group.nodes.new("NodeGroupInput")
        node_transform = node_group.nodes.new("GeometryNodeMeshToPoints")
        node_mat       = node_group.nodes.new("GeometryNodeSetMaterial")
        node_out       = node_group.nodes.new("NodeGroupOutput")

        node_group.links.new(node_in.outputs[0], node_transform.inputs[0])
        node_group.links.new(node_transform.outputs[0], node_mat.inputs[0])
        node_group.links.new(node_mat.outputs[0], node_out.inputs[0])

        node_transform.inputs["Radius"].default_value = mesh_cfg.wireframe_tubes_radius * mesh_cfg.wireframe_tubes_vertex_radius_mult
        node_mat.inputs["Material"].default_value = wireframe_mat

        modifier.node_group = node_group

    # If we are using our custom tessellator, apply it after all of the 
    # wireframe logic above, and replace the mesh which we actually render
    # with one tesselated by our algorithm
    if mesh_cfg.triangulate_polygons_with == 'BLENDER':
        # this is default behavior
        pass

    elif mesh_cfg.triangulate_polygons_with == 'MESHGEN_TOOLS':
        # remove the old mesh object
        bpy.data.objects.remove(mesh)

        # create a new mesh object
        mesh = read_mesh(global_cfg, mesh_cfg, global_translate, global_scale, return_verts_and_faces=False, use_triangulated=True)

    else:
        raise ValueError(f"Unrecognized value for mesh_cfg.triangulate_polygons_with: {mesh_cfg.triangulate_polygons_with}")


    ## Apply appearance settings to the main mesh 

    # apply shading
    if mesh_cfg.shade_smooth:
        mesh.data.shade_smooth()
    else:
        mesh.data.shade_flat()
   
    # Select what kind of material to apply to the mesh
    if mesh_cfg.scalar_func_source != None:
        mesh_mat, objects = create_uvscalar_material(mesh_cfg, f"{mesh_cfg.name}-MeshUVScalarColorMat")
        created_objects += objects

    else:
        mesh_mat = create_material(resolve_color(mesh_cfg.color), f"{mesh_cfg.name}-MeshMat")

    assign_mesh_material(mesh, mesh_mat)

    if not mesh_cfg.show_mesh:
        bpy.data.objects.remove(mesh)

    created_objects.append(mesh)

    return created_objects

def configure_point_cloud_in_scene(global_cfg, point_cloud_cfg):

    # global transforms
    # (local transforms are stored inside mesh_cfg and separately applied internally in read_mesh())
    global_translate = global_cfg.scene.global_transform_translate
    s = global_cfg.scene.global_transform_scale
    global_scale = (s, s, s) 

    ## read main mesh
    mesh = read_mesh(global_cfg, point_cloud_cfg, global_translate, global_scale)

    # make sure the names in the config and the blender object match
    # (if no name was given in the yml, use Blender's auto-assigned name)
    if 'name' not in point_cloud_cfg:
        point_cloud_cfg.name = mesh.name
    else:
        mesh.name = point_cloud_cfg.name
    point_cloud_cfg.name = f"{point_cloud_cfg.name}-PointCloud"
    mesh.name = point_cloud_cfg.name

    point_mat = create_material(resolve_color(point_cloud_cfg.color), f"{point_cloud_cfg.name}-Wireframe-MeshMat")


    ## Add a Geometry Node modifier to create points

    modifier = mesh.modifiers.new("ScriptAdded", "NODES")

    node_group = bpy.data.node_groups.new("group name", type='GeometryNodeTree')
    node_group.interface.new_socket(name="Geometry In", in_out ="INPUT", socket_type="NodeSocketGeometry")
    node_group.interface.new_socket(name="Geometry Out", in_out ="OUTPUT", socket_type="NodeSocketGeometry")

    node_in        = node_group.nodes.new("NodeGroupInput")
    node_transform = node_group.nodes.new("GeometryNodeMeshToPoints")
    node_mat       = node_group.nodes.new("GeometryNodeSetMaterial")
    node_out       = node_group.nodes.new("NodeGroupOutput")

    node_group.links.new(node_in.outputs[0], node_transform.inputs[0])
    node_group.links.new(node_transform.outputs[0], node_mat.inputs[0])
    node_group.links.new(node_mat.outputs[0], node_out.inputs[0])

    node_transform.inputs["Radius"].default_value = point_cloud_cfg.radius
    node_mat.inputs["Material"].default_value = point_mat

    modifier.node_group = node_group

    return [mesh]

def add_sun_light(rotation_euler, strength, shadow_soft_size = 0.05):
    # adapted from BlenderToolbox

    x = rotation_euler[0] * 1.0 / 180.0 * np.pi 
    y = rotation_euler[1] * 1.0 / 180.0 * np.pi 
    z = rotation_euler[2] * 1.0 / 180.0 * np.pi 
    angle = (x,y,z)

    # Create light and link
    light_data = bpy.data.lights.new(name='SUNLIGHT', type='SUN')
    light_object = bpy.data.objects.new(name="SUNLIGHT_OBJ", object_data=light_data)
    bpy.context.collection.objects.link(light_object)

    light_object.rotation_euler = angle
    light_data.angle = shadow_soft_size
    light_data.energy = strength
    # light_data.use_nodes = True
    # light_data.node_tree.nodes["Emission"].inputs['Strength'].default_value = strength

    return light_object

def add_ambient_light(strength=0.1):
    bpy.data.scenes[0].world.use_nodes = True
    bpy.data.scenes[0].world.node_tree.nodes["Background"].inputs['Color'].default_value = (1., 1., 1., 1.)
    bpy.data.scenes[0].world.node_tree.nodes["Background"].inputs['Strength'].default_value = strength

def create_shadow_threshold_composite_pass(alpha_threshold, interpolation_mode = 'CARDINAL', denoise=True):
    # adapted from BlenderToolbox

    bpy.context.scene.use_nodes = True
    tree = bpy.context.scene.node_tree
    RAMP = tree.nodes.new('CompositorNodeValToRGB')
    RAMP.color_ramp.elements[0].color[3] = 0
    RAMP.color_ramp.elements[0].position = alpha_threshold
    RAMP.color_ramp.interpolation = interpolation_mode
   
    if denoise:
        DENOISE_RGB = tree.nodes.new('CompositorNodeDenoise')
        DENOISE_ALPHA = tree.nodes.new('CompositorNodeDenoise')

    REND = tree.nodes["Render Layers"]
    OUT = tree.nodes["Composite"]

    if denoise:
        tree.links.new(REND.outputs[0], DENOISE_RGB.inputs[0])
        tree.links.new(DENOISE_RGB.outputs[0], OUT.inputs[0])

        tree.links.new(REND.outputs[1], DENOISE_ALPHA.inputs[0])
        tree.links.new(DENOISE_ALPHA.outputs[0], RAMP.inputs[0])
    else:
        tree.links.new(REND.outputs[1], RAMP.inputs[0])

    tree.links.new(RAMP.outputs[1], OUT.inputs[1])

def set_camera_look_at(camera, point):
    # adapted from BlenderToolbox
    direction = point - camera.location
    rotQuat = direction.to_track_quat('-Z', 'Y')
    camera.rotation_euler = rotQuat.to_euler()

def create_camera(location, look_at=(0,0,0), rotation_euler=None, focal_deg=40, turntable_rot=None):
    # adapted from BlenderToolbox

    if turntable_rot is not None:
        location = (np.cos(turntable_rot)*location[0] - np.sin(turntable_rot)*location[1],
                    np.sin(turntable_rot)*location[0] + np.cos(turntable_rot)*location[1],
                    location[2])

    bpy.ops.object.camera_add(location = location) # name 'Camera'
    cam = bpy.context.object
    cam.data.lens_unit = 'FOV'
    cam.data.angle = np.pi * focal_deg / 180.

    if look_at is not None:
        if rotation_euler is not None: 
            raise ValueError('camera settings must specify exactly one of `look_at` and `rotation_euler`')
        loc = mathutils.Vector(look_at)
        set_camera_look_at(cam, loc)

    if rotation_euler is not None:
        if look_at is not None: 
            raise ValueError('camera settings must specify exactly one of `look_at` and `rotation_euler`')
        cam.rotation_euler[0] = rotation_euler[0] * 1.0 / 180.0 * np.pi
        cam.rotation_euler[1] = rotation_euler[1] * 1.0 / 180.0 * np.pi
        cam.rotation_euler[2] = rotation_euler[2] * 1.0 / 180.0 * np.pi

    return cam

def create_shadowcatcher_ground(location = (0,0,0), ground_size = 100, shadow_brightness = 0.7, ground_level=None):
    # adapted from BlenderToolbox

    if ground_level is not None:
        location = (location[0], location[1], ground_level)

    # initialize a ground for shadow
    bpy.ops.mesh.primitive_plane_add(location=location, size=ground_size)
    bpy.context.object.is_shadow_catcher = True # for blender 3.X

    # # set material
    ground = bpy.context.object
    mat = bpy.data.materials.new('MeshMaterial')
    ground.data.materials.append(mat)
    mat.use_nodes = True
    tree = mat.node_tree
    tree.nodes["Principled BSDF"].inputs['Transmission Weight'].default_value = shadow_brightness

    return ground

mat_id = 0
def create_material(color, mat_name=None):
    global mat_id

    if mat_name == None:
        mat_name = f"assigned_mat_{mat_id:06d}"
        mat_id += 1

    mat = bpy.data.materials.new(mat_name)
    mat.use_nodes = True
    tree = mat.node_tree

    # set principled BSDF
    tree.nodes["Principled BSDF"].inputs['Base Color'].default_value = color
    tree.nodes["Principled BSDF"].inputs['Roughness'].default_value = 0.3
    tree.nodes["Principled BSDF"].inputs['Sheen Tint'].default_value = [0, 0, 0, 1]
    tree.nodes["Principled BSDF"].inputs['Specular IOR Level'].default_value = 0.5
    tree.nodes["Principled BSDF"].inputs['IOR'].default_value = 1.45
    tree.nodes["Principled BSDF"].inputs['Transmission Weight'].default_value = 0
    tree.nodes["Principled BSDF"].inputs['Coat Roughness'].default_value = 0

    return mat


def create_uvscalar_material(obj_cfg, mat_name=None):
    global mat_id

    # prep the colormap
    colormap_filename = save_matplotlib_colormap_as_tmp_image_file(obj_cfg.scalar_func_cmap)
    cmap_image = bpy.data.images.load(filepath=colormap_filename, check_existing=False)

    if mat_name == None:
        mat_name = f"assigned_mat_{mat_id:06d}"
        mat_id += 1

    mat = bpy.data.materials.new(mat_name)
    mat.use_nodes = True
    tree = mat.node_tree

    ## Create nodes

    uv_map_node = mat.node_tree.nodes.new('ShaderNodeUVMap')
    sep_xyz_node = mat.node_tree.nodes.new('ShaderNodeSeparateXYZ')

    map_range_node = mat.node_tree.nodes.new('ShaderNodeMapRange')
    map_range_node.inputs['From Min'].default_value = float(obj_cfg.scalar_func_vmin)
    map_range_node.inputs['From Max'].default_value = float(obj_cfg.scalar_func_vmax)

    combine_xyz_node = mat.node_tree.nodes.new('ShaderNodeCombineXYZ')
    combine_xyz_node.inputs['Y'].default_value = 0.5

    image_texture_node = mat.node_tree.nodes.new('ShaderNodeTexImage')
    image_texture_node.image = cmap_image
    image_texture_node.extension = 'EXTEND'

    default_bsdf_node = tree.nodes["Principled BSDF"]
    default_bsdf_node.inputs['Roughness'].default_value = 0.3
    default_bsdf_node.inputs['Sheen Tint'].default_value = [0, 0, 0, 1]
    default_bsdf_node.inputs['Specular IOR Level'].default_value = 0.5
    default_bsdf_node.inputs['IOR'].default_value = 1.45
    default_bsdf_node.inputs['Transmission Weight'].default_value = 0
    default_bsdf_node.inputs['Coat Roughness'].default_value = 0

    ## Hook em up
    mat.node_tree.links.new(uv_map_node.outputs['UV'], sep_xyz_node.inputs['Vector'])
    mat.node_tree.links.new(sep_xyz_node.outputs['X'], map_range_node.inputs['Value'])
    mat.node_tree.links.new(map_range_node.outputs['Result'], combine_xyz_node.inputs['X'])
    mat.node_tree.links.new(combine_xyz_node.outputs['Vector'], image_texture_node.inputs['Vector'])
    mat.node_tree.links.new(image_texture_node.outputs['Color'], default_bsdf_node.inputs['Base Color'])

    return mat, [cmap_image]

def assign_mesh_material(mesh, mat):
    mesh.data.materials.append(mat)
    mesh.active_material = mat

def composite_over_color(image_filename, color_str):

    # argument should be a hex str
    if not isinstance(color_str, str):
        raise ValueError("background color should be a string")
    if not color_str.startswith('#'): # it's hard to type #FFFFFF on the command line, so you can omit the # and we will add it
        color_str = '#' + color_str
    bg_color = ImageColor.getcolor(color_str, "RGB") # conver to a color

    composite_over(image_filename, background_color=bg_color)

def stats_to_transforms(cfg, summary_stats):

    summary_stats['scale'] = 0.5*np.max(summary_stats['bbox_max'] - summary_stats['bbox_min'])
    
    if cfg.scene.global_transform_scale == 'auto':
        cfg.scene.global_transform_scale = float(1.0/summary_stats['scale'])
        print(f"Auto-set global_transform_scale = {cfg.scene.global_transform_scale}")
    summary_stats['bbox_min'] *= cfg.scene.global_transform_scale
    summary_stats['bbox_max'] *= cfg.scene.global_transform_scale
    # global_transform_translate = np.array(cfg.scene.global_transform_translate)
    # global_transform_translate /= cfg.scene.global_transform_scale
    # cfg.scene.global_transform_translate = global_transform_translate.tolist()

    if cfg.scene.global_transform_translate == 'auto':
        center = 0.5 * (summary_stats['bbox_min'] + summary_stats['bbox_max'])
        cfg.scene.global_transform_translate = (-center).tolist()
        print(f"Auto-set global_transform_translate = {-center}")
    summary_stats['bbox_min'] += np.array(cfg.scene.global_transform_translate)
    summary_stats['bbox_max'] += np.array(cfg.scene.global_transform_translate)


    summary_stats['lowest_coord'] = summary_stats['bbox_min'][2]
    if cfg.scene.ground_level == 'auto':
        cfg.scene.ground_level = float(summary_stats['lowest_coord'])
        print(f"Placed ground at {cfg.scene.ground_level}")

    if cfg.scene.shift_lowest_content_to_origin_floor:
        global_transform_translate = np.array(cfg.scene.global_transform_translate)
        global_transform_translate[2] -= summary_stats['lowest_coord']
        cfg.scene.global_transform_translate = global_transform_translate.tolist()
        summary_stats['bbox_min'][2] -= summary_stats['lowest_coord']
        summary_stats['bbox_max'][2] -= summary_stats['lowest_coord']
        cfg.scene.ground_level -= float(summary_stats['lowest_coord'])
        summary_stats['lowest_coord'] = 0.
        print("Shifted lowest content to origin")
            


def compute_auto_transforms(cfg, batch_render_ind=None):
    # Compute transforms for all of the scene data at the specified render ind and store it on cfg

    # Annoyingly this is stateful
    if batch_render_ind is not None:
        saved_batch_render_ind = cfg.batch_render_ind
        cfg.batch_render_ind = batch_render_ind


    ## Handle automatically computed statistics, if needed
    need_stats = cfg.scene.global_transform_translate == 'auto' or \
                 cfg.scene.global_transform_scale == 'auto' or \
                 cfg.scene.ground_level == 'auto' or \
                 cfg.scene.shift_lowest_content_to_origin_floor
    if need_stats:

        print("\n=== Computing scene statistics for automatic layout")

        INF = float('inf')
        summary_stats = {
            'bbox_min' : np.array((+INF,+INF,+INF)),
            'bbox_max' : np.array((-INF,-INF,-INF)),
        }

        # Do a first pass walking through all of the objects to compute
        # bounding boxes, etc

        for content_entry in cfg.content:
            for typekey in content_entry:
                value_cfg = content_entry[typekey]

                print(f"computing scene stats for {value_cfg}")
        
                if typekey == 'mesh':
                    print(f"Computing scene stats for mesh")
                    stats = compute_mesh_scene_stats(cfg, value_cfg)

                elif typekey == 'point_cloud':
                    print(f"Computing scene stats for point cloud")
                    stats = compute_point_cloud_scene_stats(cfg, value_cfg)

                else:
                    raise ValueError(f'Unrecognized content key {typekey}')

                summary_stats['bbox_min'] = np.minimum(summary_stats['bbox_min'], stats['bbox_min'])
                summary_stats['bbox_max'] = np.maximum(summary_stats['bbox_max'], stats['bbox_max'])

        stats_to_transforms(cfg, summary_stats)

        print("  ...done computing scene statistics")
    
    # Annoyingly this is stateful
    if batch_render_ind is not None:
        cfg.batch_render_ind = saved_batch_render_ind 

def compute_shared_auto_transforms(cfg):
    
    # Compute transforms for all of the scene data at ALL specified render inds and store it on cfg

    # Annoyingly this is stateful
    saved_batch_render_ind = cfg.batch_render_ind

    ## Handle automatically computed statistics, if needed
    need_stats = cfg.scene.global_transform_translate == 'auto' or \
                 cfg.scene.global_transform_scale == 'auto' or \
                 cfg.scene.ground_level == 'auto' or \
                 cfg.scene.shift_lowest_content_to_origin_floor
    if need_stats:

        print("\n=== Computing scene statistics for automatic layout")

        INF = float('inf')
        summary_stats = {
            'bbox_min' : np.array((+INF,+INF,+INF)),
            'bbox_max' : np.array((-INF,-INF,-INF)),
        }

        # Do a first pass walking through all of the objects to compute
        # bounding boxes, etc

        for i_batch in range(cfg.batch_render_lim):

            cfg.batch_render_ind = i_batch

            for content_entry in cfg.content:
                for typekey in content_entry:
                    value_cfg = content_entry[typekey]

                    print(f"computing scene stats for {value_cfg}")
            
                    if typekey == 'mesh':
                        print(f"Computing scene stats for mesh")
                        stats = compute_mesh_scene_stats(cfg, value_cfg)

                    elif typekey == 'point_cloud':
                        print(f"Computing scene stats for point cloud")
                        stats = compute_point_cloud_scene_stats(cfg, value_cfg)

                    else:
                        raise ValueError(f'Unrecognized content key {typekey}')

                    summary_stats['bbox_min'] = np.minimum(summary_stats['bbox_min'], stats['bbox_min'])
                    summary_stats['bbox_max'] = np.maximum(summary_stats['bbox_max'], stats['bbox_max'])

        stats_to_transforms(cfg, summary_stats)

        print("  ...done computing scene statistics")
    
    # Annoyingly this is stateful
    cfg.batch_render_ind = saved_batch_render_ind 

def main():

    parser = argparse.ArgumentParser()

    parser.add_argument("-c", "--configs", nargs='+', type=str, help="Paths to one or more configuration files, which will be applied in sequence on top of the base config")
    parser.add_argument("-o", "--overrides", type=str, nargs='*', default=[], help="Individual overrides for configuration settings")

    args = parser.parse_args()

    # Load the default config
    cfg = OmegaConf.load(DEFAULT_CONFIG_PATH)

    if args.configs is not None:
        for file in args.configs:
            file_config = OmegaConf.load(file)
            cfg = OmegaConf.merge(cfg, config)
    
    # Apply command-line overrides
    config_overrides = OmegaConf.from_dotlist(args.overrides)
    cfg = OmegaConf.merge(cfg, config_overrides)

    ## Add a few of our own properties to the global config
    cfg = OmegaConf.merge(OmegaConf.create({
                            'is_batch_rendering' : False,
                            'batch_render_ind' : 0, 
                            'batch_render_lim' : 'auto',
                        }), cfg)

    print("Final configuration:")
    print(OmegaConf.to_yaml(cfg))

    ## Initialize blender
    initialize_blender(cfg)

    ## Prep the content list

    # If a mesh/point_cloud/etc entry is given at the top level,
    # add it to the content list to be processed
    if 'mesh' in cfg:
        cfg.content.append({'mesh' : cfg.mesh})
    if 'point_cloud' in cfg:
        cfg.content.append({'point_cloud' : cfg.point_cloud})
   
    if len(cfg.content) == 0:
        raise ValueError("no content has been added to the scene!")

    ## Apply default configs
    # Walk through the list of passed objects in the content list,
    # for each, identify it as a mesh/point_cloud/etc, and merge in
    # with the appropriate config
    # TODO I think this can be done nicer with structured configs
    for ic in range(len(cfg.content)):
        content_entry = cfg.content[ic]
        for typekey in content_entry:
            value = content_entry[typekey]
            if typekey == 'mesh':
                value = OmegaConf.merge(DEFAULT_MESH_CONFIG, value)
            elif typekey == 'point_cloud':
                value = OmegaConf.merge(DEFAULT_POINT_CLOUD_CONFIG, value)
            else:
                raise ValueError(f'Unrecognized content key {typekey}')

            content_entry[typekey] = value # updating dict while iterating is safe

        # replace the updated entry into the content list
        cfg.content[ic] = content_entry

    # Handle batch rendering auto-detection
    # If any of the input content are files, count the number of matching files
    batch_render_count = 1
    for content_entry in cfg.content:
        for typekey in content_entry:
            value_cfg = content_entry[typekey]

            if typekey == 'mesh':
                _, n_batch = resolve_possible_batch_rendering_path(cfg, value_cfg)

            elif typekey == 'point_cloud':
                _, n_batch = resolve_possible_batch_rendering_path(cfg, value_cfg)

            else:
                raise ValueError(f'Unrecognized content key {typekey}')

            if n_batch != 1:
                if batch_render_count == 1:
                    batch_render_count = n_batch
                elif batch_render_count != n_batch:
                    raise ValueError(f'Got multiple content batches for batch rendering, but they are of different size: {batch_render_count} and {n_batch}')

    if cfg.batch_render_lim == 'auto':
        cfg.batch_render_lim = batch_render_count
    if cfg.batch_render_lim > 1:
        cfg.is_batch_rendering = True
        cfg.batch_render_lim = min(cfg.batch_render_lim, batch_render_count)

    # Do some validation and settings management
    if cfg.output.movie and not cfg.is_batch_rendering:
        raise ValueError('Got movie=True, but did not recieve a batch of files as input. Make sure at least one of the input args points at a directory')
    if cfg.output.movie and cfg.render.transparent:
        raise ValueError("Transparency and movies don't play nice together, set render.transparent=False")
    if cfg.output.movie:
        cfg.scene.share_auto_transforms_across_batch = True
    

    ## Precompute transforms if needed
    if cfg.scene.share_auto_transforms_across_batch:
        compute_shared_auto_transforms(cfg)

    ## Add per-scene data which never changes

    # Add a compositing step to threshold grey shadows to completely white
    if cfg.scene.with_ground_shadows and cfg.render.threshold_ground_shadows:
        # if we have ground shadows, denoising is done in the compositor rather than the renderer
        create_shadow_threshold_composite_pass(alpha_threshold=cfg.render.threshold_ground_shadows_alpha, denoise=cfg.render.denoise)

    # Scene content which may be created below
    cam = None
    shadow_catcher = None
    sun_light = None
    ambient_light = None

    ## Walk through the iterations and render each
    # (in an ordinary one-off render which is not batched, this loop just iterates once)
    frame_count = max(cfg.batch_render_lim, cfg.render.animate_frames)
    for i_batch in range(frame_count):

        cfg.batch_render_ind = i_batch

        if not cfg.scene.share_auto_transforms_across_batch:
            compute_auto_transforms(cfg)


        ## Walk through all of the content in the scene and actually add it
        content_objects = []
        for content_entry in cfg.content:
            for typekey in content_entry:
                value_cfg = content_entry[typekey]

                if typekey == 'mesh':
                    content_objects += configure_mesh_in_scene(cfg, value_cfg)

                elif typekey == 'point_cloud':
                    content_objects += configure_point_cloud_in_scene(cfg, value_cfg)

                else:
                    raise ValueError(f'Unrecognized content key {typekey}')


        ## Create a shadow catcher
        if cfg.scene.with_ground_shadows:
            if shadow_catcher is not None:
                bpy.data.objects.remove(shadow_catcher)
            shadow_catcher = create_shadowcatcher_ground(shadow_brightness=0.9, ground_level=cfg.scene.ground_level)

        ## Set the camera
        if cam is not None:
            delete_object(cam)
            
        # set up turntable rotation
        turntable_rot = 0.
        if cfg.render.animate_frames > 1 and cfg.render.animate_camera == 'turntable':
            turntable_rot = 2. * np.pi * cfg.batch_render_ind / float(cfg.render.animate_camera_turntable_period)


        cam = create_camera(location=tuple(cfg.view.camera_location), 
                            look_at=tuple(cfg.view.camera_look_at) if cfg.view.camera_look_at is not None else None, 
                            rotation_euler=tuple(cfg.view.camera_rotation_euler) if cfg.view.camera_rotation_euler is not None else None, 
                            focal_deg=cfg.view.focal_deg,
                            turntable_rot=turntable_rot
                            )

        
        ## Add lights
        if sun_light is not None:
            delete_object(sun_light)
        sun_light = add_sun_light(cfg.scene.sun_light_euler_angle, cfg.scene.sun_light_strength, cfg.scene.sun_light_shadow_softness)
        if ambient_light is not None:
            delete_object(ambient_light)
        ambient_light = add_ambient_light(cfg.scene.ambient_light_strength)

        ## save blender file so that you can debug or adjust parameters in the UI
        if cfg.output.save_scene:
            ensure_dir_exists(TMP_DIR)
            bpy.ops.wm.save_mainfile(filepath=os.path.join(TMP_DIR, 'tmp_renderscene.blend'))

        if cfg.is_batch_rendering or cfg.render.animate_frames > 1:
            directory, base, extension = split_filepath(cfg.output.image_filename)
            outname = os.path.join(directory, f"{base}_{cfg.batch_render_ind:06d}{extension}")
        else:
            outname = cfg.output.image_filename

        ## Render the actual image
        bpy.data.scenes['Scene'].render.filepath = outname
        bpy.data.scenes['Scene'].camera = cam 
        bpy.ops.render.render(write_still = True)

        # If a non-transparent image was requested, composite it over the background color
        if not cfg.render.transparent:
            composite_over_color(outname, cfg.render.background_color)

        # Clean up
        for ob in content_objects:
            delete_object(ob)

if __name__ == "__main__":
    main()
